from setuptools import find_packages, setup

setup(
    name="{{ workspace_name }}",
    version="0.0.1",
    packages=find_packages(),
    install_requires=[
        "dagster",
        "dagster-cloud",
        "dagster-sdf",
    ],
    extras_require={
        "dev": [
            "dagster-webserver",
        ]
    },
)
